;; "GALAXY collisions" using Greenspan 1st-order discrete mechanics
;; with vector operations for motion of massless tracer particles ("stars").
;; In later comments, "bodies" are the galaxy cores, whose motion is
;; calculated in the host, also using 1st-order discrete mechanics.
;; This version uses fractal "star clouds" for bright arms, randomly
;; dimmed stars between arms.  Intended for use with depth-sensitive
;; plotting.
;; 
;; Updated version 5/1/91 now calculates and stores "body" positions in 
;; the host before calculating any tracer particle moves.  This is done
;; in the function "rung1"

(in-package '*lisp)

(defconstant NBQ 401)       ; Maximum number of saved galaxy cores
(defconstant NBX 4)         ; Maximum number of galaxies, allows for
                            ; messages of 32*NBX bits
(defconstant D2R 57.29578)  ; Degrees per radian

(defvar plotting NIL)       ; Set to TERM or COLOR by plotter files

(*proclaim '(type fixnum N LN CN NB NP PX NX NA saveX))

(defvar CN 12)  ; Field size for grid and raster coordinates
(defvar NB  0)  ; Actual number of bodies
(defvar NX  0)  ; Number of tracer particles
(defvar PX  2)  ; Print/plot control
(defvar CG  0)  ; Current galaxy core timestep
(defvar ITRS  0); Number of iterations till stars converge in Green1
(defvar saveX 0); save stars for this galaxy core timestep

;; Geometry for VP set, calculated by setup

(defvar GEOM nil)

(*proclaim '(type single-float G TWOPI dt hdt simtim))

(defvar G  0.0)             ; Gravitational constant
(defvar TWOPI 6.2831853)
(defvar dt  0.0)            ; Time step, delta-t
(defvar hdt 0.0)            ; Half delta-t
(defvar simtim 0.0)         ; Simulation time

;; Make an n-element vector

(defun nvec (n)
 (declare (type integer n))
 (make-array (list n) :element-type 'single-float :initial-element 0.0)
)

;; Make simple host array of n 3-element, single-float, simple vectors

(defun 3darray (n)
 (declare (type integer n))
 (let ((m (make-array (list n))))
  (dotimes (i n) (setf (aref m i) (nvec 3)))
  m
 )
)

;; Make host array of n timesteps of 4-galaxy arrays

(defun gqueue (n)
 (declare (type integer n))
 (let ((m (make-array (list n))))
  (dotimes (i n) (setf (aref m i) (3darray NBX)))
  m
 )
)

(*proclaim '(type (vector t) gx0 gv0 gv1 gac gwv))

(defvar gx0 (gqueue NBQ))     ; Galaxy positions for galaxy core timesteps
(defvar gv0 (3darray NBX))    ; velocities
(defvar gv1 (3darray NBX))    ;
(defvar gac (3darray NBX))    ; accelerations
(defvar gwv (3darray NBX))    ; working vector


(*proclaim '(type (vector single-float) gm mj gcd))

(defvar gm (nvec NBX))   ; Masses
(defvar mj (nvec NBX))   ; Scaled masses
(defvar gcd (nvec NBX))  ; Galactic core diameter

(*proclaim '(type (array fixnum (* 3)) grgb))

(defvar grgb (make-array (list NBX 3)
 :element-type 'fixnum :initial-element 0))  ; rgb values for galaxies

(proclaim '(type (vector single-float 3) gww gwu gwr))

(defvar gww #(0.0 0.0 0.0)) ; Working host vectors
(defvar gwu #(0.0 0.0 0.0))
(defvar gwr #(0.0 0.0 0.0))


(*proclaim
  '(type (vector-pvar single-float 3) x0 x1 v0 v1 d0 d1 ac wv xhold vhold))

(*defvar x0 (!! #(0.0 0.0 0.0)))    ; positions of tracer particles ("stars")
(*defvar x1 (!! #(0.0 0.0 0.0)))
(*defvar v0 (!! #(0.0 0.0 0.0)))    ; velocities
(*defvar v1 (!! #(0.0 0.0 0.0)))
(*defvar d0 (!! #(0.0 0.0 0.0)))    ; displacements
(*defvar d1 (!! #(0.0 0.0 0.0)))
(*defvar ac (!! #(0.0 0.0 0.0)))    ; accelerations
(*defvar wv (!! #(0.0 0.0 0.0)))    ; working vector
(*defvar xhold (!! #(0.0 0.0 0.0))) ; saved tracer particles of galaxy saveX
(*defvar vhold (!! #(0.0 0.0 0.0))) ;  

(*proclaim '(type single-float-pvar r0 r1 w))

(*defvar r0 (!! 0.0))    ; distances
(*defvar r1 (!! 0.0))
(*defvar w  (!! 0.0))    ; working variable

(*proclaim '(type (unsigned-pvar *log-number-of-processors-limit*) self))

(*defvar self (self-address!!))  ; working self-address

(*proclaim '(type (unsigned-pvar 8) prd pgr pbl))

(*defvar prd (!! 0))             ; Red value for "star"
(*defvar pgr (!! 0))             ; Green do.
(*defvar pbl (!! 0))             ; Blue  do.

(*proclaim '(type boolean-pvar act))

(*defvar act  nil!!)  ; active in "star space"

(*proclaim '(type fixnum fri frj grx gry glx gly))

(defvar frx 0)                 ;; Counter for fractal selection
(defvar fri 0)  (defvar frj 0) ;; Select fractal maps for current galaxy
(defvar grx 1)  (defvar gry 2) ;; Grid dimensions in blocks of 128x32
(defvar glx 128)               ;; Grid size in coord 0
(defvar gly 128)               ;; Grid size in coord 1

;; Parameters for fractal representations, darkening

(*proclaim '(type single-float gfra gfrb gfrc gbas gfrh))
(*proclaim '(type fixnum gdrk))

(defvar gfra 0.60) (defvar gfrb 0.82) (defvar gfrc 0.95)
(defvar gbas 4.50) (defvar gfrh 0.83)
(defvar gdrk 300)  ; Min brightness in dark areas, on scale of 1024

;; Cold-boot and setup.  Cm need not be cold-booted; uses default if
;; Cm not attached.

(defun setup (&optional (sx 128) (sy 128))
 (*cold-boot :initial-dimensions (list sx sy))
 (setq NX 0)
 (if (eq plotting 'COLOR) (gfrctl gfra gfrb gfrc gbas gfrh))
 (setq glx (car *current-cm-configuration*))
 (setq gly (cadr *current-cm-configuration*))
 (setq grx (ash 1 (- (dimension-address-length 0) 7)))
 (setq gry (ash 1 (- (dimension-address-length 1) 5)))
)

;; Construct galaxy number ig with ns stars, radii 0 < rc < ri < ro,
;; center gx, velocity gv, rotated by phi around axis at theta,
;; color given by rd, gr, bl.  The galaxy has 4 spiral arms.
;; Uses random darkening between arms, fractal brightness in arms

(defun init-gxy (ig ns rc ri ro ma gx gv theta phi rd gr bl)
 (declare (type fixnum ig rd gr bl) (type (vector single-float 3) gx gv)
          (type integer ns) (type single-float rc ri ro ma theta phi))
 (c3v (aref (aref gx0 cg) ig) gx)  (c3v (aref gv0 ig) gv)
 (setf (aref gm ig) ma)  (setf (aref mj ig) (* G ma))
 (setf (aref gcd ig) rc)
 (setf (aref grgb ig 0) rd)
 (setf (aref grgb ig 1) gr)
 (setf (aref grgb ig 2) bl)
 (setq fri (* 128 (mod ig grx)))
 (setq frj (* 32 (mod (truncate ig grx) gry)))
 (mkgxy ri ro ns ma gx gv theta phi rd gr bl)
)

;; Construct galaxy disk with radii 0 < ri < ro, n stars, central mass ma.
;; Galaxy centered at gx with velocity gv (vectors), oriented by rotation
;; of phi around axis at theta

(*defun mkgxy (ri ro n ma gx gv theta phi rd gr bl)
 (declare (type single-float ri ro ma theta phi) (type fixnum rd gr bl)
          (type (vector single-float 3) gx gv) (type integer n))
 (if (> (+ NX n) *number-of-processors-limit*)
  (format t "~%Too few processors to continue.~%")
  (let* ((ro-ri (- ro ri))
         (ro+ri (+ ro ri))
         (k  (floor (sqrt (/ (* n ro-ri) (* pi ro+ri)))))
         (k1 (1+ k))
         (n0 (round (/ (* 2 ri n) (* ro+ri k1))))
         (dn (ceiling (/ (* ro-ri n0) (* ri k)))))
        (declare (type single-float ro-ri ro+ri) (type integer k n0 dn))
;   (format t "k = ~D, n0 = ~D, dn = ~D~%" k n0 dn)
   (*let ((i (-!! self (!! NX))) (j (!! 0)) b c fb (fj (!! 0))
          rj x y s vx vy sn cs)
         (declare (type (signed-pvar 24) i j b c fb fj)
                  (type single-float-pvar rj x y s vx vy sn cs))
    (*when (<!! (!! -1) i (!! n))
     (*set c (!! k1)) (*set act t!!)
     ;; inv:  f(j) = fj <= i & j < c & (i > f(c) or c = k+1),
     ;;       where f(j) = j*n0 + dn*(j-1)*j/2 (no. of stars in rings < j)
     (dotimes (m (integer-length k1))
      (*when (>!! (-!! c j) (!! 1))
        (*set b (ash!! (+!! j c) (!! -1)))  ;; j < b < c
        (*set fb (+!! (*!! b (!! n0))
                      (*!! (!! dn) (ash!! (*!! (1-!! b) b) (!! -1)))))
        (*if (<=!! fb i)
             (progn (*set j b) (*set fj fb))  (*set c b))  ;; end *if
        )) ;; end *when, end dotimes
    (*set i (-!! i fj)) ;; position in ring j
    (*if (<!! j (!! k)) (*set b (+!! (!! n0) (*!! j (!! dn))))
                        (*set b (-!! (!! n) fj)))
         ;; b is number of stars in this ring
    (if (eq plotting 'COLOR) (spiral-arms))
    (*set rj (+!! (!! ri) (*!! j (!! (/ ro-ri k))))) ;; radius of this ring
    (*set s (sqrt!! (/!! (!! (* G ma)) rj)))    ;; circular speed
    (*set x (/!! (*!! (!! TWOPI) i) b))         ;; orientation of this star
    (*set sn (sin!! x))   (*set cs (cos!! x))   ;; unit vector to this star
    (*set x (*!! rj cs))  (*set y (*!! rj sn))  ;; position in x-y plane
    (*set vx (-!! (*!! s sn))) (*set vy (*!! s cs))  ;; velocity in x-y plane
    (let* ((sth (sin theta)) (cth (cos theta))
           (sph (sin phi))   (cph (cos phi))
           (sn2 (* sth sth)) (cs2 (* cth cth))
           (ax  (+ cs2 (* sn2 cth))) (ay (* sth cth (- 1.0 cph))))
          (declare (type single-float sth cth sph cph sn2 cs2 ax ay))
      ;; Rotate disk and move to correct position
      (*setf (aref!! x0 (!! 0)) (+!! (*!! (!! ax)  x) (*!! (!! ay)  y)))
      (*setf (aref!! v0 (!! 0)) (+!! (*!! (!! ax) vx) (*!! (!! ay) vy)))
      (setq ax (* sth cth (- 1.0 cph)))  (setq ay (+ sn2 (* cs2 cph)))
      (*setf (aref!! x0 (!! 1)) (+!! (*!! (!! ax)  x) (*!! (!! ay)  y)))
      (*setf (aref!! v0 (!! 1)) (+!! (*!! (!! ax) vx) (*!! (!! ay) vy)))
      (setq ax (- (* sth sph))) (setq ay (* cth sph))
      (*setf (aref!! x0 (!! 2)) (+!! (*!! (!! ax)  x) (*!! (!! ay)  y)))
      (*setf (aref!! v0 (!! 2)) (+!! (*!! (!! ax) vx) (*!! (!! ay) vy)))
      ;; Translate and add velocities
      (*set x0 (v+!! x0 (!! gx)))  (*set v0 (v+!! v0 (!! gv)))
   )))  ;; end let*, end *when, end *let
    (setq NX (+ NX n))
    (incf frx) ; next fractal display
    (setq fri (* 128 (mod frx grx)))
    (setq frj (*  32 (mod (truncate frx grx) gry)))
 )) ;; end let*, end if
)  ;; end mkgxy

;; Calculate colors and intensities of stars, creating spiral arms.
;; This macro is intended to be called only at the appropriate place
;; in mkgxy.

(defmacro spiral-arms ()
 '(*let ((lf (!! 0.0)))
        (declare (type single-float-pvar lf))
    (*set prd (!! rd))  (*set pgr (!! gr)) (*set pbl (!! bl))
    (*set c (mod!! (+!! i j j j) b))  ;; position in spiral structure
    (*if (oddp!! (truncate!! (*!! (!! 8) c) b)) (darken) (make-clouds))
    (*set prd (min!! (!! 255) (round!! (*!! prd lf))))
    (*set pgr (min!! (!! 255) (round!! (*!! pgr lf))))
    (*set pbl (min!! (!! 255) (round!! (*!! pbl lf)))))
)


;; Darken sections between arms.  This macro is intended to be called only
;; at the appropriate place in spiral-arms.

(defmacro darken ()
  '(*set lf (/!! (+!! (random!! (!! (- 1024 gdrk))) (!! gdrk)) (!! 1024.0)))
)

;; Make clouds within arms.  This macro is intended to be called only
;; at the appropriate 
(defmacro make-clouds ()
  '(*let* ((arm (ash!! (truncate!! (*!! (!! 4) c) b) (!! 1)))
           (fj  (round!! (*!! (!! 32) j) (!! k)))
           (c0  (ash!! (+!! (*!! arm b) (!! 7)) (!! -3)))
           (c1  (ash!! (1-!! (*!! (1+!! arm) b)) (!! -3)))
           (dc  (-!! c1 c0))
           (fi  (round!! (*!! (!! 32) (-!! c c0)) dc))
          )
          (declare (type (signed-pvar  8) arm fj fi))
          (declare (type (signed-pvar 24) c0 c1 dc))
         (*set lf
	    (pref!! frctlx
               (grid!! (mod!! (+!! (!! fri) (ash!! arm (!! 4)) fi) (!! glx))
                       (mod!! (+!! (!! frj) fj) (!! gly)))))
    ) ; end *let*
)

;; Load  gravitational const, initial positions, velocities, masses,
;; construct galaxies.  Argument is stream from which to read.

(defun init(strm)
  (setf G (read strm))
  (setf NB (read strm))
  (dotimes (k NB)
   (let ((gx (nvec 3)) (gv (nvec 3)) m)
    (dotimes (u 3) (setf (aref gx u) (read strm)))
    (dotimes (u 3) (setf (aref gv u) (read strm)))
    (setq m (setf (aref gm k) (read strm)))
    (init-gxy k (read strm) (read strm) (read strm) (read strm) m gx gv
           (/  (read strm) D2R) (/  (read strm) D2R)
           (read strm) (read strm) (read strm))
   ))  ; end let, end dotimes
)

(defun print-state(np)  ;; Print galaxy positions
 (format t
 "~%       x         y         z         vx        vy        vz        m~%")
 (dotimes (k np)
  (let ((gx (aref (aref gx0 cg) k)) (gv (aref gv0 k)))
   (format t "~2D ~9,5F ~9,5F ~9,5F ~9,5F ~9,5F ~9,5F ~9,3E~%"
     k (aref gx 0) (aref gx 1) (aref gx 2) 
       (aref gv 0) (aref gv 1) (aref gv 2)
       (aref gm k))))
)

;; Print np entries of vector-pvar x, selected uniformly across stars

(*defun nbfltpr (x np)
 (let ((str (truncate NX np)) (ucb 0))
   (dotimes (u np)
     (setq ucb (* u str))
     (format t "~6D " ucb)
     (dotimes (v 3)
       (format t "~11,4E " (pref (aref!! x (!! v)) ucb)))
     (terpri)
   ))  ; end dotimes, end let
)

;; Copy 3-vector

(defun c3v (dst src)
 (declare (type (vector single-float 3) dst src))
 (dotimes (i 3) (setf (aref dst i) (aref src i)))
)

;; Copy n by 3 array (as for galaxy data)

(defun cnx3 (dst src)
 (declare (type (vector t) dst src))
 (dotimes (ig NB)
   (let ((dv (svref dst ig)) (sv (svref src ig)))
        (declare (type (vector single-float 3) dv sv))
    (dotimes (i 3) (setf (aref dv i) (aref sv i)))
   )) ; end let, end dotimes
)

;; Calculate max abs value of components of 3-vector

(defun max3 (v)
 (declare (type (vector single-float 3) v))
 (max (abs (svref v 0)) (abs (svref v 1)) (abs (svref v 2)))
)

;; Accumulate into 3-vector

(defun vac3 (dst src)
 (declare (type (vector single-float 3) dst src))
 (incf (aref dst 0) (aref src 0))
 (incf (aref dst 1) (aref src 1))
 (incf (aref dst 2) (aref src 2))
)

;; Calculate galaxy core accelerations      

(defun gaccel ()
 (dotimes (ig NB) (c3v (aref gac ig) #(0.0 0.0 0.0)))
 (dotimes (ig (1- NB))
  (let ((ix0 (aref (aref gx0 cg) ig)) (ix1 (aref (aref gx0 (1+ cg)) ig)) 
        (iac (aref gac ig)) (im (aref mj ig)))
       (declare (type (vector single-float 3) ix0 ix1 iac)
                      (type single-float im))
   (do ((jg (1+ ig) (incf jg))) ((>= jg NB))
    (let* ((jx0 (aref (aref gx0 cg) jg))  (jx1 (aref (aref gx0 (1+ cg)) jg))  
           (jac (aref gac jg))
           (d0 (v- ix0 jx0))    (d1 (v- ix1 jx1))    (jm  (aref mj jg))
           (r0 (vabs d0))       (r1 (vabs d1))
           (acv (v/-constant (v+ d1 d0) (* r0 r1 (+ r0 r1)))))
          (declare (type (vector single-float 3) jx0 jx1 jac d0 d1 acv)
                   (type single-float jm r0 r1))
      (vac3 iac (v- (v*-constant acv jm)))  (vac3 jac (v*-constant acv im))
    )) ; end let*, end do
  )) ; end let, end dotimes
)

(proclaim '(fixnum cfail chfail))
(defvar cfail 0)    ; Counts number of iterations that stars failed to converge
(defvar chfail 0)   ; Counts number of iterations that cores failed to converge

;; Advance galaxy cores one timestep. This is performed entirely in the host.
;; eps is solution tolerance, lim is max number of iterations.  Returns
;; actual number of iterations and tallies convergence failures in chfail.

(defun greenh (eps lim)
 (declare (single-float eps) (fixnum lim))
 (let ((ic 0) (mx 0.0))
      (declare (fixnum ic) (single-float mx))
   (tagbody
     again (incf ic)
	   (gaccel)  ; Calculate galaxy accelerations
           (dotimes (ig NB) ; Update max
             (setq gwu (aref gv1 ig))
	     (setq gwr (aref gv0 ig))
             (setq gww (v+ gwr (v*-constant (aref gac ig) dt)))
             (setq mx (max3 (v- gww gwu)))
             (c3v gwu gww)
             (setq gww (v+ (aref (aref gx0 cg) ig)
			   (v*-constant (v+ gwu gwr) hdt)))
             (setq gwr (aref (aref gx0 (1+ cg)) ig))
             (setq mx (max mx (max3 (v- gww gwr))))
             (c3v gwr gww)
           ) ; end dotimes
       (if (or (< mx eps) (= ic lim)) (go finished))
	     (go again)
         finished
    ) ; end tagbody
    (if (= ic lim) (incf chfail))
  ic) ; end let
)

;; Calculate accelerations on stars

(defun accel ()
 (let ((cx0 (aref gx0 cg)) (cx1 (aref gx0 (1+ cg))))
  (*set ac (!! #(0.0 0.0 0.0)))
  (dotimes (ig NB)
    (*set d0 (v-!! (!! (the (vector single-float 3) (aref cx0 ig))) x0))
    (*set r0 (vabs!! d0))
    (*set d1 (v-!! (!! (the (vector single-float 3) (aref cx1 ig))) x1))
    (*set r1 (vabs!! d1))
    (*set wv (v+!! d1 d0))
    (*set wv (v*scalar!! wv (!! (the single-float (aref mj ig)))))
    (*set w (*!! r0 r1 (+!! r0 r1)))
    (*set wv (v/scalar!! wv w))
    (*set ac (v+!! ac wv))
   ) ; end (dotimes ...)
 ) ; end (let ...)
)

;; Calculate max absolute value of vector-pvar components
;; Written with defun rather than *defun because of compiler problems

(defun max3!! (x)
 (declare (type (vector-pvar single-float 3) x))
 (*set w (max!! (abs!! (aref!! x (!! 0)))
                (abs!! (aref!! x (!! 1)))
	        (abs!! (aref!! x (!! 2)))))
 (*max w)  ;; Result
)
   

;; Greenspan's 1st-order method.  Performs ns steps of time dt
;; with solution tolerance eps.  Returns number of evaluations.
;; Leaves number of iterations which were stopped at limit in cfail.

(defun green1 (ns eps)
 (declare (fixnum ns) (single-float eps))
 (let ((ic 0) (v 0) (lim 10) (mx 0.0))
      (declare (fixnum ic v lim) (single-float mx))
  (setf hdt (* 0.5 dt))
  (*set x1  x0)   (*set v1  v0)  ; initialize for star integration
  (*when act
    (dotimes (k ns)
      (setq ic 0)
      (tagbody
	again (incf v) (incf ic)
              (accel)
              (*set wv (v*scalar!! ac (!! dt)))
	      (*set wv (v+!! v0 wv))
	      (*set d0 (v-!! wv v1))
	      (setf mx (max3!! d0))
	      (*set v1 wv)
              (*set wv (v+!! v0 v1))
              (*set wv (v*scalar!! wv (!! hdt)))
	      (*set wv (v+!! x0 wv))
              (*set d1 (v-!! wv x1))
	      (setf mx (max mx (max3!! d1)))
	      (*set x1 wv)
	 (if (or (< mx eps) (= ic lim)) (go finished))
	      (go again)
         finished
       ) ; end tagbody
      (when (eq saveX cg) (*set xhold x0) (*set vhold v0))
      (*set x0 x1) (*set v0 v1)      ; Update star positions, velocities
      (*set v1 (v+!! v0 ac))
      (*set x1 (v+!! x0 (v*scalar!! (v+!! v0 v1) (!! hdt))))
      (if (= ic lim) (incf cfail))
      (incf cg)                      ; Increment to next core timestep
    ) ; end dotimes
   ) ; end *when
   (setq ITRS v)
  v) ; end let
)

;; Simulated time

(proclaim '(type single-float simtm))

(defvar simtm 0.0)

;; Print  or plot current state of simulation, according to PX

(defun gx-pstate ()
 (if (<= PX 2) (print-state NB) (plot-galaxies))
)

;; This calculates the first ns*np galaxy core timesteps in the host and
;; saves them in the array gx0.
;; Run integration sequence ns time steps per cycle, np cycles of output
;; time step dt, tolerance eps

(defun rung1 (ns np xdt eps)
 (declare (fixnum ns np) (single-float xdt eps))
 (setq cg 0)                               ;; initialize galaxy core timestep
 (let ((hv 0) (lim 10))
  (setf dt xdt)  (setf hdt (* 0.5 dt))
  (cnx3 (aref gx0 (+ cg 1)) (aref gx0 cg))  ;; init. for galaxy core integr.
  (cnx3 gv1 gv0)                            ;; init. for galaxy core integr.
  (dotimes (j np)
   (dotimes (k ns)
    (incf hv (greenh eps lim))
    (incf cg)
    (cnx3 (aref gx0 (+ 1 cg)) (aref gx0 cg)) ;; update galaxy info.
    (cnx3 gv0 gv1)                           ;; update vel. for galaxies
   ) ; end dotimes (k ns)
  ) ; end dotimes (j np)
  (setq cg 0)                                ;; initialize galaxy timestep
 )
)

;; This calculates and plots the "star" positions for the galaxies.
;; Run integration sequence ns timesteps per cycle, np cycles of output
;; time step dt, tolerance eps
;; NB: This should be run with same xdt and eps of rung1 and optimally
;;     with same ns and np.

(defun rung2 (ns np xdt eps)
 (declare (fixnum ns np) (single-float xdt eps))
 (setq cg 0 simtm 0.0)                       ;; initialize galaxy timestep
 (format t "~%   G = ~13,6E  delta-t = ~12,5F  epsilon = ~12,5E~%"
                        G                dt                eps     )
 (format t "   ~D galaxies~%" NB)
 (format t "~% time = ~F" simtm) 
 (gx-pstate)
 (cntg2 ns np xdt eps)
)

;; Continue star integration from previous time step, parameters as
;; for rung2 above

(defun cntg2 (ns np xdt eps)
 (declare (fixnum ns np) (single-float xdt eps))
 (setf dt xdt)  (setf hdt (* 0.5 dt))
 (let ((vl 0))
  (dotimes (k np)
    (setf cfail 0)
    (incf simtm (* ns dt))
    (setf vl (green1 ns eps))
    (if (< PX 3)
      (format t 
         "~% simulated time = ~F  ~D evaluations  ~D iterations limited~%"
                             simtm vl            cfail))
    (gx-pstate)
   ) ; end dotimes
  (if (= PX 3) (format t "~% simulated time = ~F~%" simtm))
 )
)

;; Same as rung2 but faster output due to removing terminal output.

(defun rung2fast (ns np xdt eps)
 (declare (fixnum ns np) (single-float xdt eps))
 (setq cg 0 simtm 0.0)                     ;; initialize galaxy timestep
 (gx-pstate)
 (cntg2fast ns np xdt eps)
)

;; Continue rung2fast from current state

(defun cntg2fast (ns np xdt eps)
 (declare (fixnum ns np) (single-float xdt eps))
  (setf dt xdt)  (setf hdt (* 0.5 dt))
  (gx-pstate)
  (dotimes (k np)
    (green1 ns eps)
    (gx-pstate)
   ) ; end dotimes
)


;; Calculates galaxy core positions and then immediately calculates and
;; plots star positions.

(defun grung (ns np xdt eps)
 (declare (fixnum ns np) (single-float xdt eps))
  (rung1 ns np xdt eps)
  (rung2 ns np xdt eps)
)

;; Same as grung but runs at optimal speed since we're calling rung2fast

(defun grungfast (ns np xdt eps)
 (declare (fixnum ns np) (single-float xdt eps))
  (rung1 ns np xdt eps)
  (rung2fast ns np xdt eps)
)

;; Recalculates and plots star positions from saved galaxy core "saveX".
;; This is initialized when the file is loaded to saveX = 0.  
;; YOU SHOULD ONLY RUN THIS AFTER RUNNING rung1 FIRST!!!

(defun rerung (ns np eps)
 (declare (fixnum ns np) (single-float eps))
 (let ((vl 0))
  (setq cg saveX)
  (setq simtm (* saveX dt))
  (*set x0 xhold)
  (*set v0 vhold)
  (format t "~%   G = ~13,6E  delta-t = ~12,5F  epsilon = ~12,5E~%"
                  G dt eps)
  (format t "   ~D galaxies~%" NB)
  (format t "~% time = ~F" simtm) 
  (gx-pstate)
  (dotimes (k np)
   (setf cfail 0)
   (incf simtm (* ns dt))
   (setf vl (green1 ns eps))
   (if (or (< PX 3) (eq plotting 'COLOR))
     (format t 
         "~% simulated time = ~F  ~D evaluations  ~D iterations limited~%"
                              simtm vl cfail))
   (gx-pstate)
  )
 )
)

;; Same as rerung but is optimal speed due to no screen output

(defun rerungfast (ns np eps)
 (declare (fixnum ns np) (single-float eps))
  (setq cg saveX)
  (*set x0 xhold)
  (*set v0 vhold)
  (gx-pstate)
  (dotimes (k np)
   (green1 ns eps)
   (gx-pstate)
  )
)

;; Manage calculation

(defun main ()
  (let (ns np cmd dt eps file strm)
    (setf PX 2)
    (tagbody
      again (format t "galaxy> ") (setf cmd (read))
	(case cmd
	  (q (go finished))
	  (p (setf PX (read)))
	  (c (grung ns np dt eps))
	  (r (go run))
	  (T (if (listp cmd) (eval cmd)
                             (format t "Unrecognized command~%"))))
	(go again)
     run
	(format t "Data file> ")  (setf file (read-line))
	(setf strm (open file :direction :input))
	(setup)
	(setf NX 0)  (init strm)
	(setf ns (read strm))  (setf np  (read strm))
	(setf dt (read strm))  (setf eps (read strm))
	(close strm)
	(grung ns np dt eps)
	(go again)
     finished)
  )
)
